import os

import pytest

from mario.data_extractor import DataExtractor, Configuration
from mario.dataset_specification import dataset_from_json
from mario.metadata import metadata_from_json, Item
from mario.validation import DataFrameValidator, HyperValidator, Validator, SqlValidator


def get_validator(nulls=False, hyperfile=False):
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    # postal code has NULLs
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    filename = 'orders_with_nulls.hyper' if nulls else 'orders.hyper'
    hyper_path = os.path.join('test', filename)

    configuration = Configuration(
        file_path=hyper_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame()
    )
    hyper_validator = HyperValidator(
        dataset_specification=dataset,
        metadata=metadata,
        hyper_file_path=hyper_path
    )
    if hyperfile:
        return hyper_validator
    return validator


def test_no_nulls():
    validator = get_validator()
    assert validator.validate_data(allow_nulls=False)
    assert len(validator.errors) == 0


def test_no_nulls_hyper():
    validator = get_validator(hyperfile=True)
    assert validator.validate_data(allow_nulls=False)
    assert len(validator.errors) == 0


def test_nulls():
    validator = get_validator(nulls=True)
    validator.validate_data(allow_nulls=True)
    assert len(validator.errors) == 0
    assert "Validation warning: 'Postal Code' contains NULLs" in validator.warnings

    with pytest.raises(ValueError):
        validator.validate_data(allow_nulls=False)
    assert len(validator.errors) == 1
    assert "Validation error: 'Postal Code' contains NULLs" in validator.errors


def test_nulls_hyper():
    validator = get_validator(nulls=True, hyperfile=True)
    validator.validate_data(allow_nulls=True)
    assert len(validator.errors) == 0
    assert "Validation warning: 'Postal Code' contains NULLs" in validator.warnings

    with pytest.raises(ValueError):
        validator.validate_data(allow_nulls=False)
    assert len(validator.errors) == 1
    assert "Validation error: 'Postal Code' contains NULLs" in validator.errors


def test_missing_column():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.dimensions.append('bigliness')
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    item = Item()
    item.name='bigliness'
    metadata.add_item(item)
    hyper_path = os.path.join('test', 'orders.hyper')

    configuration = Configuration(
        file_path=hyper_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )

    with pytest.raises(ValueError):
        validator.validate_data(allow_nulls=True)
    assert len(validator.errors) == 1
    assert validator.errors[0] == "Validation error: 'bigliness' in specification is missing from dataset"


def test_missing_column_hyper():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.dimensions.append('bigliness')
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    item = Item()
    item.name = 'bigliness'
    metadata.add_item(item)
    hyper_path = os.path.join('test', 'orders.hyper')

    hyper_validator = HyperValidator(
        dataset_specification=dataset,
        metadata=metadata,
        hyper_file_path=hyper_path
    )
    with pytest.raises(ValueError):
        hyper_validator.validate_data(allow_nulls=True)
    assert len(hyper_validator.errors) == 1
    assert hyper_validator.errors[0] == "Validation error: 'bigliness' in specification is missing from dataset"

# TODO
# Range checks

def test_domain_check():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    item = metadata.get_metadata('Ship Mode')
    item.set_property('domain', ["First Class", "Second Class", "Standard Class"])
    hyper_path = os.path.join('test', 'orders.hyper')

    configuration = Configuration(
        file_path=hyper_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )

    with pytest.raises(ValueError):
        validator.validate_data(allow_nulls=True)
    assert len(validator.errors) == 1
    assert validator.errors[0] == "Validation error: 'Same Day' is not in domain of 'Ship Mode'"


def test_domain_check_hyper():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    item = metadata.get_metadata('Ship Mode')
    item.set_property('domain', ["First Class", "Second Class", "Standard Class"])
    hyper_path = os.path.join('test', 'orders.hyper')

    validator = HyperValidator(
        dataset_specification=dataset,
        metadata=metadata,
        hyper_file_path=hyper_path
    )

    with pytest.raises(ValueError):
        validator.validate_data(allow_nulls=True)
    assert len(validator.errors) == 1
    assert validator.errors[0] == "Validation error: 'Same Day' is not in domain of 'Ship Mode'"


def test_range_check():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    item = metadata.get_metadata('Discount')
    item.set_property('range', [0.0, 0.2])
    hyper_path = os.path.join('test', 'orders.hyper')

    configuration = Configuration(
        file_path=hyper_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )

    with pytest.raises(ValueError):
        validator.validate_data(allow_nulls=True)
    assert len(validator.errors) == 1
    assert validator.errors[0] == "Validation error: 'Discount': '0.8' is greater than '0.2'"


def test_range_check_hyper():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    item = metadata.get_metadata('Discount')
    item.set_property('range', [0.0, 0.2])
    hyper_path = os.path.join('test', 'orders.hyper')

    validator = HyperValidator(
        dataset_specification=dataset,
        metadata=metadata,
        hyper_file_path=hyper_path
    )

    with pytest.raises(ValueError):
        validator.validate_data(allow_nulls=True)
    assert len(validator.errors) == 1
    assert validator.errors[0] == "Validation error: 'Discount': '0.8' is greater than '0.2'"


def test_multiple_errors_hyper():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    item = metadata.get_metadata('Discount')
    item.set_property('range', [0.0, 0.2])
    item = metadata.get_metadata('Ship Mode')
    item.set_property('domain', ["First Class", "Second Class", "Standard Class"])
    hyper_path = os.path.join('test', 'orders_with_nulls.hyper')

    validator = HyperValidator(
        dataset_specification=dataset,
        metadata=metadata,
        hyper_file_path=hyper_path
    )

    with pytest.raises(ValueError):
        validator.validate_data(allow_nulls=False)
    assert len(validator.errors) == 3
    assert "Validation error: 'Discount': '0.8' is greater than '0.2'" in validator.errors
    assert "Validation error: 'Same Day' is not in domain of 'Ship Mode'" in validator.errors
    assert "Validation error: 'Postal Code' contains NULLs" in validator.errors


def test_pattern_validation_passes():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    dataset.dimensions.append('Order Identifier')
    file_path = os.path.join('test', 'orders.csv')

    configuration = Configuration(
        file_path=file_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )

    validator.validate_data(allow_nulls=True)
    assert len(validator.errors) == 0


def test_pattern_validation_fails():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    dataset.dimensions.append('Order Identifier')
    item = metadata.get_metadata('Order Identifier')
    item.set_property('pattern', 'US-20\d\d-\d{6}')
    file_path = os.path.join('test', 'orders.csv')

    configuration = Configuration(
        file_path=file_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )

    with pytest.raises(ValueError):
        validator.validate_data(allow_nulls=True)
    assert len(validator.errors) == 102
    assert "Validation error: 'Order Identifier': 'CA-2019-115238' does not match the pattern 'US-20\\d\\d-\\d{6}'" in validator.errors


def test_get_hierarchy_data():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    file_path = os.path.join('test', 'orders.csv')

    configuration = Configuration(
        file_path=file_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )

    dimensions = validator.__get_data_for_hierarchy__('Product')
    assert len(dimensions) == 1849
    assert len(dimensions.columns) == 2


def test_check_hierarchies():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    file_path = os.path.join('test', 'orders.csv')

    configuration = Configuration(
        file_path=file_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )
    validator.check_hierarchies()

    assert validator.errors == ["Inconsistent hierarchy: 92024 at level Postal Code is represented in multiple higher level categories ('United States', 'West', 'California', 'Encinitas') and ('United States', 'West', 'California', 'San Diego')."]


def test_check_hierarchies_hyper():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    file_path = os.path.join('test', 'orders.hyper')

    validator = HyperValidator(
        dataset_specification=dataset,
        metadata=metadata,
        hyper_file_path=file_path
    )
    validator.check_hierarchies()

    assert validator.errors == ["Inconsistent hierarchy: 92024 at level Postal Code is represented in multiple higher level categories ('United States', 'West', 'California', 'Encinitas') and ('United States', 'West', 'California', 'San Diego')."]


def test_category_anomalies():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    file_path = os.path.join('test', 'orders.csv')

    configuration = Configuration(
        file_path=file_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    # Introduce a segmentation variable
    df = extractor.get_data_frame()
    df['Year'] = df['Ship Date'].str[-4:]

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )
    validator.check_category_anomalies('Year')

    assert "Validation warning: 'Ship Mode' has potentially anomalous data when segmented by 'Year'" in validator.warnings


def test_category_anomalies_hyper():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    file_path = os.path.join('test', 'orders.hyper')

    configuration = Configuration(
        file_path=file_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    # Introduce a segmentation variable
    df = extractor.get_data_frame()
    df['Year'] = df['Ship Date'].astype(str).str[0:4]

    output_file_path = os.path.join('output', 'orders_with_segmentation.hyper')
    extractor.save_data_as_hyper(file_path=output_file_path, minimise=False)

    validator = HyperValidator(
        dataset_specification=dataset,
        metadata=metadata,
        hyper_file_path=output_file_path
    )
    validator.check_category_anomalies('Year')

    assert "Validation warning: 'Ship Mode' has potentially anomalous data when segmented by 'Year'" in validator.warnings


def test_all_checks():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    file_path = os.path.join('test', 'orders.csv')

    configuration = Configuration(
        file_path=file_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    # Introduce a segmentation variable
    df = extractor.get_data_frame()
    df['Year'] = df['Ship Date'].str[-4:]

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )
    with pytest.raises(ValueError):
        validator.validate_data(check_hierarchies=True, detect_anomalies=True, segmentation='Year')
    assert len(validator.errors) == 1
    assert len(validator.warnings) == 11


def test_all_checks_sql():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))

    configuration = Configuration(
        view='superstore',
        schema='dev',
        connection_string=os.environ.get("CONNECTION_STRING")
    )

    validator = SqlValidator(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    with pytest.raises(ValueError):
        validator.validate_data(check_hierarchies=True, detect_anomalies=False)

    assert len(validator.errors) == 2
    assert len(validator.warnings) == 7


def test_checks_iteratively():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    file_path = os.path.join('test', 'orders.csv')

    configuration = Configuration(
        file_path=file_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )
    for item in dataset.items:
        validator.errors = []
        validator.warnings = []
        validator.validate_data_item(item, allow_nulls=False)
        assert validator.errors == []
        if item == 'Ship Date':
            assert len(validator.warnings) == 2
        elif item in ['Country/Region', 'State/Province', 'City', 'Product Name', 'Sales', 'Profit']:
            assert len(validator.warnings) == 1
        else:
            assert len(validator.warnings) == 0


def test_check_anomalies_single_field():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    file_path = os.path.join('test', 'orders.csv')

    configuration = Configuration(
        file_path=file_path
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )

    # Introduce a segmentation variable
    df = extractor.get_data_frame()
    df['Year'] = df['Ship Date'].str[-4:]

    validator = DataFrameValidator(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor.get_data_frame(minimise=False)
    )
    validator.check_item_for_anomalies('Ship Mode', 'Year')
    assert "Validation warning: 'Ship Mode' has potentially anomalous data when segmented by 'Year'" in validator.warnings




from setuptools import setup

setup(
    name='mario-pipeline-tools',
    version='0.48',
    packages=['mario'],
    url='https://github.com/JiscDACT/mario',
    license='all rights reserved',
    author='scottbw',
    author_email='scott.wilson@jisc.ac.uk',
    description='Base classes and helpers for common data pipeline tasks',
    long_description='A set of base classes and helpers for common data pipeline tasks, '
                     'whether using in local control scripts or in Airflow tasks.',
    package_data={
        # If any package contains *.ini files, include them
        '': ['*.ini']
    },
    include_package_data=True,
    install_requires=[
        'pandas',
        'pypika',
        'sqlalchemy',
        'openpyxl'
    ],
    extras_require={
        'Airflow': ['apache-airflow-providers-common-sql'],
        'Tableau': ['pantab', 'tableauhyperapi', 'tableau-builder==0.18']
    }
)


import os
import tempfile

from mario.metadata import metadata_from_json, metadata_from_excel


def test_load_metadata():
    metadata_file = os.path.join('test', 'metadata.json')
    metadata = metadata_from_json(file_path=metadata_file)
    assert metadata.get_metadata('Ship Mode') is not None
    assert metadata.get_metadata('Ship Mode').description is not None
    assert metadata.get_metadata('Ship Mode').get_property('groups') == ['Shipping']
    assert metadata.get_property('description') is not None


def test_save_metadata():
    metadata_file = os.path.join('test', 'metadata.json')
    metadata = metadata_from_json(file_path=metadata_file)
    metadata.set_property("fruit", "banana")
    with tempfile.NamedTemporaryFile() as file:
        file.close()
        metadata.save(file_path=file.name)
        metadata = metadata_from_json(file.name)
        assert metadata.get_property('fruit') == 'banana'


def test_load_tdsa():
    metadata_file = os.path.join('test', 'tdsa.json')
    metadata = metadata_from_json(file_path=metadata_file)
    assert metadata.get_metadata('Ship Mode') is not None


def test_load_excel():
    metadata_file = os.path.join('test', 'spec_example.xlsx')
    metadata = metadata_from_excel(file_path=metadata_file)
    domain = metadata.get_metadata('Nationality (UK/ EU/ Non-EU/ Unknown) (2022/23 onwards)').get_property('domain')
    assert 'UK' in domain
    assert '23 onwards' not in domain
    assert metadata.get_metadata('ShipMode (First Class / Same Day / Second Class/Standard Class)') is not None
    assert metadata.get_metadata('Region') is not None
    assert metadata.get_metadata('ShipMode (First Class / Same Day / Second Class/Standard Class)').get_property('domain') is not None
    domain = metadata.get_metadata('ShipMode (First Class / Same Day / Second Class/Standard Class)').get_property('domain')
    assert 'First Class' in domain


def test_get_hierarchies():
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    hierarchies = metadata.get_hierarchies()
    assert len(hierarchies) == 2
    assert 'Product' in hierarchies
    assert 'Location' in hierarchies


def test_get_hierarchy():
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    products = metadata.get_hierarchy('Product')
    assert products == ['Category', 'Product Name']


import logging
from copy import copy
from typing import List

from pypika import Table, Criterion, PostgreSQLQuery as Query, functions as fn, Parameter

from mario.data_extractor import Configuration
from mario.dataset_specification import DatasetSpecification
from mario.metadata import Metadata

logger = logging.getLogger(__name__)


def get_formatted_query(query, params):
    formatted_query = copy(query)
    for key, value in params.items():
        formatted_query = formatted_query.replace(f'%({key})s', f"'{value}'")
    return formatted_query


class QueryBuilder:
    """
    Generic interface for a query builder. Subclass this to create your own custom
    query builders to pass to a DataExtractor in a Configuration.
    """
    def __init__(self,
                 configuration: Configuration,
                 metadata: Metadata,
                 dataset_specification: DatasetSpecification):
        self.configuration = configuration
        self.metadata = metadata
        self.dataset_specification = dataset_specification

    def create_query(self) -> [str, List[any]]:
        raise NotImplementedError

    def create_totals_query(self, measure=None) -> [str, List[any]]:
        raise NotImplementedError


class ViewBasedQueryBuilder(QueryBuilder):
    """
    Creates a very simple 'get everything from the view' query
    """
    def __init__(self,
                 dataset_specification: DatasetSpecification,
                 configuration: Configuration,
                 metadata: Metadata):
        super().__init__(
            configuration=configuration,
            dataset_specification=dataset_specification,
            metadata=metadata
        )

    def create_totals_query(self, measure=None) -> [str, List[any]]:
        if measure is None:
            _sql = f'SELECT COUNT(*) FROM "' + self.configuration.schema + '"."' + self.configuration.view + '"'
        else:
            _sql = f'SELECT SUM("'+measure+'") FROM "' + self.configuration.schema + '"."' + self.configuration.view + '"'

        _params = []
        return [_sql, _params]

    def create_query(self) -> [str, List[any]]:
        _sql = 'SELECT * FROM "' + self.configuration.schema + '"."' + self.configuration.view + '"'
        _params = []
        return [_sql, _params]


class SubsetQueryBuilder(QueryBuilder):
    """
    Class for building SQL queries as selections from a view with aggregation
    """
    def __init__(self,
                 dataset_specification: DatasetSpecification,
                 configuration: Configuration,
                 metadata: Metadata):
        super().__init__(
            configuration=configuration,
            dataset_specification=dataset_specification,
            metadata=metadata
        )
        self.table = Table(self.configuration.view, schema=self.configuration.schema)
        self.years = self.dataset_specification.get_property('years')
        self.onward_use_category = self.dataset_specification.get_property('onwardUseCategory')
        logger.debug("Query builder using table " + str(self.table))

    def contains_subject(self) -> bool:
        for item in self.dataset_specification.items:
            meta = self.metadata.get_metadata(item)
            if meta.get_property('groups') is not None:
                if 'subject' in meta.get_property('groups'):
                    return True
        return False

    def create_totals_query(self, measure=None) -> [str, List[any]]:
        """
        Constructs a 'total only' query by combining the selections and constraints.
        :return: an array containing the query prepared statement, and the parameters
        TODO make this more generic
        """
        measures = []
        if measure is None:
            for measure in self.dataset_specification.measures:
                # Decide column name based on whether 'subject' fields are present
                if measure in ['FPE', 'FTE'] and not self.contains_subject():
                    measures.append(fn.Sum(self.table[measure], 'Count'))
                else:
                    measures.append(fn.Sum(self.table[measure], measure))
        else:
            measures.append(fn.Sum(self.table[measure], measure))

        q = Query().from_(self.table).select(*measures)

        q, parameters = self.create_constraints(q)
        return [q.get_sql(), parameters]

    def create_query(self) -> [str, List[any]]:
        """
        Constructs the query by combining the selections and constraints.
        :return: an array containing the query prepared statement, and the parameters
        """

        select_fields = []
        for field in self.dataset_specification.items:
            # Don't include calculated fields
            meta = self.metadata.get_metadata(field)
            if not meta.get_property('formula'):
                select_fields.append(field)
        group_fields = select_fields.copy()

        # remove measures from regular select
        measures = []
        for measure in self.dataset_specification.measures:
            if measure in select_fields:
                select_fields.remove(measure)
                group_fields.remove(measure)
                # Decide column name based on whether 'subject' fields are present
                if measure in ['FPE', 'FTE'] and not self.contains_subject():
                    measures.append(fn.Sum(self.table[measure], 'Count'))
                else:
                    measures.append(fn.Sum(self.table[measure], measure))
        select_fields.extend(measures)

        q = Query(). \
            from_(self.table). \
            select(*select_fields). \
            groupby(*group_fields)

        q, parameters = self.create_constraints(q)
        return [q.get_sql(), parameters]

    def create_constraints(self, q):
        """
        Builds a set of constraints from the query; this consists of both
        defined constraints, and the implicit constraint on years. Uses the
        'in' method for determining constraint syntax
        :param q: the Query object
        :return: the Query object with constraints
        """
        clauses = []
        parameters = {}
        for constraint in self.dataset_specification.constraints:
            column = constraint.item
            placeholders = []
            for i in range(len(constraint.allowed_values)):
                parameter_name = column.replace(" ", "_") + str(i)
                # Postgres style.
                # TODO Probably need some way of identifying which parameter style to apply.
                parameter = Parameter('%('+parameter_name+')s')
                placeholders.append(parameter)
                parameters[parameter_name] = constraint.allowed_values[i]
            clauses.append(self.table[column].isin(placeholders))

        q = q.where(Criterion.all(clauses))
        return q, parameters
